use crate::vm::dm_type::DmType;
use crate::vm::dvm_value::DvmValue;
use crate::vm::object_type::{DmField, DmImplementation};
use std::alloc::{alloc, dealloc, Layout};
use std::collections::HashSet;
use std::rc::Rc;

#[derive(Debug, PartialEq, Eq)]
pub struct DmAllocObject {
    pub data: *mut u8,
    pub units: HashSet<usize>,
    pub size: usize,
    pub layout: Layout,
    pub implementation: Rc<DmImplementation>,
}

impl DmAllocObject {
    pub fn new(size: usize, implementation: Rc<DmImplementation>) -> Self {
        let layout = Layout::from_size_align(size, 1).unwrap();
        DmAllocObject {
            data: unsafe { alloc(layout) },
            units: HashSet::new(),
            size,
            layout,
            implementation,
        }
    }
}

impl Drop for DmAllocObject {
    fn drop(&mut self) {
        unsafe {
            dealloc(self.data, self.layout);
        }
    }
}

pub unsafe fn get_field_value(dm_field: &DmField, self_object: &DmAllocObject) -> DvmValue {
    let data_size = dm_field.dm_type().size_in_bytes();
    let mut raw_data: Vec<u8> = Vec::with_capacity(data_size);
    for i in dm_field.data_offset()..(dm_field.data_offset() + data_size) {
        raw_data.push(self_object.data.offset(i as isize).read());
    }
    match dm_field.dm_type() {
        DmType::Byte => DvmValue::Byte(raw_data[0]),
        DmType::Int => DvmValue::Int(i32::from_ne_bytes(
            raw_data[0..data_size].try_into().unwrap(),
        )),
        DmType::Long => DvmValue::Long(i64::from_ne_bytes(
            raw_data[0..data_size].try_into().unwrap(),
        )),
        DmType::Double => DvmValue::Double(f64::from_ne_bytes(
            raw_data[0..data_size].try_into().unwrap(),
        )),
        DmType::Boolean => DvmValue::Boolean(raw_data[0] != 0),
        DmType::Pointer => {
            // read the pointer's (address) value
            let address = usize::from_ne_bytes(raw_data[0..data_size].try_into().unwrap());
            DvmValue::Pointer(Rc::from_raw(address as *const DmAllocObject))
        }
        DmType::Unit => DvmValue::Uninit,
    }
}

macro_rules! read_array {
    ( $T: ty, $raw_data: expr, $length: expr ) => {{
        let t_size = size_of::<$T>();
        let mut arr: Vec<$T> = Vec::with_capacity($length);
        for i in 0..$length {
            arr.push(<$T>::from_ne_bytes(
                $raw_data[(i * t_size)..(i * t_size + t_size)]
                    .try_into()
                    .unwrap(),
            ))
        }
        arr
    }};
}

fn read_i32_array(raw_data: &Vec<u8>, length: usize) -> Vec<i32> {
    read_array!(i32, raw_data, length)
}

fn read_i64_array(raw_data: &Vec<u8>, length: usize) -> Vec<i64> {
    read_array!(i64, raw_data, length)
}

fn read_f64_array(raw_data: &Vec<u8>, length: usize) -> Vec<f64> {
    read_array!(f64, raw_data, length)
}

fn read_bool_array(raw_data: &Vec<u8>) -> Vec<bool> {
    raw_data.iter().map(|b| *b == 1).collect()
}

fn read_usize_array(raw_data: &Vec<u8>, length: usize) -> Vec<usize> {
    read_array!(usize, raw_data, length)
}
