use crate::vm::dm_type::DmType;
use crate::vm::mem::DmAllocObject;
use std::fmt::Debug;
use std::rc::Rc;

#[derive(Debug, PartialEq, Eq)]
pub enum DmObjectType {
    Interface(DmInterface),
    Implementation(DmImplementation),
}

#[derive(Debug, Eq)]
pub struct DmFn {
    fqn: String,
    short_name: String,
    byte_code: Vec<u8>,
    number_used_registers: usize,
    return_register: Option<u8>,
}

impl DmFn {
    pub fn new(
        fqn: &str,
        short_name: &str,
        byte_code: Vec<u8>,
        number_used_registers: usize,
        return_register: Option<u8>,
    ) -> Self {
        DmFn {
            fqn: fqn.to_string(),
            short_name: short_name.to_string(),
            byte_code,
            number_used_registers,
            return_register,
        }
    }

    pub fn fqn(&self) -> &str {
        self.fqn.as_str()
    }

    pub fn short_name(&self) -> &str {
        self.short_name.as_str()
    }

    pub fn byte_code(&self) -> &Vec<u8> {
        &self.byte_code
    }

    pub fn number_used_registers(&self) -> usize {
        self.number_used_registers
    }

    pub fn return_register(&self) -> Option<u8> {
        self.return_register
    }
}

impl PartialEq for DmFn {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

#[derive(Debug, Eq)]
pub struct DmMethod {
    dm_fn: Rc<DmFn>,
    implements: Option<Rc<DmVirtualMethod>>,
}

impl PartialEq for DmMethod {
    fn eq(&self, other: &Self) -> bool {
        self.dm_fn == other.dm_fn
    }
}

impl DmMethod {
    pub fn new(dm_fn: DmFn, implements: Option<Rc<DmVirtualMethod>>) -> Self {
        DmMethod {
            dm_fn: Rc::new(dm_fn),
            implements,
        }
    }

    pub fn dm_fn(&self) -> &DmFn {
        &self.dm_fn
    }
}

#[derive(Debug, Eq)]
pub struct DmInterface {
    fqn: String,
    short_name: String,
    functions: Vec<Rc<DmFn>>,
    virtual_methods: Vec<Rc<DmVirtualMethod>>,
}

impl PartialEq for DmInterface {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmInterface {
    pub fn new(fqn: &str, short_name: &str) -> Self {
        DmInterface {
            fqn: fqn.to_string(),
            short_name: short_name.to_string(),
            functions: Vec::new(),
            virtual_methods: Vec::new(),
        }
    }

    pub fn fqn(&self) -> &str {
        self.fqn.as_str()
    }

    pub fn short_name(&self) -> &str {
        self.short_name.as_str()
    }

    pub fn add_function(&mut self, dm_fn: DmFn) {
        self.functions.push(Rc::new(dm_fn));
    }

    pub fn get_functions(&self) -> Vec<Rc<DmFn>> {
        self.functions.clone()
    }

    pub fn add_method(&mut self, dm_virtual_method: DmVirtualMethod) {
        self.virtual_methods.push(Rc::new(dm_virtual_method));
    }

    pub fn get_method(&self, name: &str, self_object: &DmAllocObject) -> Option<Rc<DmMethod>> {
        if self
            .virtual_methods
            .iter()
            .find(|&dm_fn| dm_fn.fqn == name)
            .is_some()
        {
            return self_object.implementation.get_method(name, self_object);
        }
        None
    }

    pub fn get_virtual_methods(&self) -> Vec<Rc<DmVirtualMethod>> {
        self.virtual_methods.clone()
    }

    pub fn get_property(&self, name: &str, self_object: &DmAllocObject) -> Option<&DmProperty> {
        todo!()
    }
}

#[derive(Debug, Eq)]
pub struct DmVirtualMethod {
    pub fqn: String,
    pub short_name: String,
}

impl PartialEq for DmVirtualMethod {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

#[derive(Debug, Eq)]
pub struct DmImplementation {
    pub fqn: String,
    pub short_name: String,
    pub interface: Option<Rc<DmInterface>>,
    pub functions: Vec<Rc<DmFn>>,
    pub methods: Vec<Rc<DmMethod>>,
    pub fields: Vec<DmField>,
}

impl PartialEq for DmImplementation {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmImplementation {
    pub fn new(fqn: &str, short_name: &str, interface: Option<Rc<DmInterface>>) -> Self {
        DmImplementation {
            fqn: fqn.to_string(),
            short_name: short_name.to_string(),
            interface,
            functions: Vec::new(),
            methods: Vec::new(),
            fields: Vec::new(),
        }
    }

    pub fn interface(&self) -> Option<Rc<DmInterface>> {
        self.interface.clone()
    }

    pub fn get_method(&self, name: &str, self_object: &DmAllocObject) -> Option<Rc<DmMethod>> {
        for method in &self.methods {
            if method.dm_fn.fqn == name {
                return Some(method.clone());
            } else if let Some(implements) = &method.implements {
                if implements.fqn == name {
                    return Some(method.clone());
                }
            }
        }
        None
    }

    pub fn get_property(&self, name: &str, self_object: &DmAllocObject) -> Option<Rc<DmProperty>> {
        todo!()
    }

    pub fn get_field(&self, name: &str, self_object: &DmAllocObject) -> Option<&DmField> {
        self.fields.iter().find(|field| field.name == name)
    }

    pub fn size_in_bytes(&self) -> usize {
        self.fields.iter().map(|field| field.size_in_bytes()).sum()
    }
}

#[derive(Debug, Eq)]
pub struct DmProperty {
    pub name: String,
    pub data_offset: usize,
    pub primitive_type: Rc<DmType>,
}

impl PartialEq for DmProperty {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}

#[derive(Debug, Eq)]
pub struct DmField {
    name: String,
    dm_type: DmType,
    data_offset: usize,
}

impl PartialEq for DmField {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}

impl DmField {
    pub fn new(name: &str, dm_type: DmType, data_offset: usize) -> Self {
        DmField {
            name: name.to_string(),
            dm_type,
            data_offset,
        }
    }

    pub fn name(&self) -> &str {
        &self.name
    }

    pub fn dm_type(&self) -> &DmType {
        &self.dm_type
    }

    pub fn data_offset(&self) -> usize {
        self.data_offset
    }

    pub fn size_in_bytes(&self) -> usize {
        self.dm_type.size_in_bytes()
    }
}
