use crate::vm::dvm_value::DvmValue;
use crate::vm::mem::{get_field_value, DmAllocObject};
use crate::vm::{DvmContext, DvmState};
use std::rc::Rc;

pub fn dm_print(args: Vec<DvmValue>, _state: &mut DvmState, _context: &DvmContext) -> DvmValue {
    if args.len() != 1 {
        return DvmValue::Void; // TODO: make exception
    }
    unsafe {
        print!("{}", get_string(&args[0]));
    }
    DvmValue::Void
}

pub fn dm_println(args: Vec<DvmValue>, _state: &mut DvmState, _context: &DvmContext) -> DvmValue {
    if args.len() != 1 {
        return DvmValue::Uninit;
    }
    unsafe {
        println!("{}", get_string(&args[0]));
    }
    DvmValue::Void
}

unsafe fn get_string(dvm_value: &DvmValue) -> String {
    match dvm_value {
        DvmValue::Byte(b) => b.to_string(),
        DvmValue::Int(i) => i.to_string(),
        DvmValue::Long(l) => l.to_string(),
        DvmValue::Double(d) => d.to_string(),
        DvmValue::Boolean(b) => b.to_string(),
        DvmValue::Pointer(object) => convert_to_string(object.clone()),
        DvmValue::Uninit => String::from("Uninit"),
        DvmValue::Void => String::from("Void"),
    }
}

fn convert_to_string(alloc_object_rc: Rc<DmAllocObject>) -> String {
    if alloc_object_rc.implementation.fqn == "std::core::String" {
        extract_string_from_string(alloc_object_rc.clone())
    } else {
        todo!("what happens if we don't have a String?")
    }
}

fn extract_string_from_string(string_object: Rc<DmAllocObject>) -> String {
    let bytes_field = string_object
        .implementation
        .get_field("bytes", &string_object)
        .expect("Could not get String.bytes field.");
    let bytes_object = unsafe { get_field_value(&bytes_field, &string_object) }.expect_pointer();
    if bytes_object.implementation.fqn != "std::core::ArrayImpl" {
        panic!("String.bytes field is not a std::core::ArrayImpl");
    }

    let pointer_field = bytes_object
        .implementation
        .get_field("pointer", &bytes_object)
        .expect("Could not get ArrayImpl.pointer field.");
    let pointer_object = unsafe { get_field_value(&pointer_field, &bytes_object) }.expect_pointer();
    if pointer_object.implementation.fqn != "std::unsafe::PointerImpl" {
        panic!("ArrayImpl.pointer is not a std::unsafe::PointerImpl");
    }

    let raw_address_field = pointer_object
        .implementation
        .get_field("raw_address", &pointer_object)
        .expect("Could not get PointerImpl.raw_address field.");
    let raw_address = unsafe { get_field_value(&raw_address_field, &pointer_object) }.expect_long();

    let size_field = pointer_object
        .implementation
        .get_field("size", &pointer_object)
        .expect("Could not get PointerImpl.size field.");
    let size = unsafe { get_field_value(&size_field, &pointer_object) }.expect_long();

    let raw_bytes_pointer = raw_address as usize as *mut u8;
    let mut v: Vec<u8> = Vec::new();
    for i in 0..(size as isize) {
        v.push(unsafe { raw_bytes_pointer.offset(i).read() });
    }
    String::from_utf8(v).unwrap()
}
